/* Debugging functionality. Contains several methods by which to write
   various levels of debugging output with printf. This means that debugging
   prints can be written everywhere in the code when debugging (at various
   levels) and just left there. By setting the debug level to 0 none of the
   debug info will be set. These will be in operation only if the DEBUG macro
   is defined. 

   Author: Kristoffer Lawson, Fishpool Creations Oy Ltd (setok@fishpool.com)
 
   $Header: /home/sleber/cvs/liboil/debug.h,v 1.1.1.1 2003/01/06 02:09:05 sleber Exp $ */


#ifndef DEBUG_H
#define DEBUG_H

#include "baseIncludes.h"


#ifdef DEBUG

/* Debug printf macros. All of them are of the form DnPRINTF where n is a
   level from 1 to 5. Each printf prints if db_level is greater or equal to
   n. The format and other parameters to the printf macros must be passed
   inside brackets because of macro logics. 
   Output is printed to the external FILE pointer variable 'db_fp', which is
   initialized to stderr. This variable (and db_level) can be freely
   changed to whatever the programmer wishes. A return ('\n') is added to the 
   end of the debug string.

   F.ex. D1PRINTF(("main loop iteration: %d", n));.

   Debug levels (higher numbers include the output from lower ones):

   1 - Top-level info. This is mainly used for giving information about
   what the program is doing (large loops, functions and other blocks),
   possibly including any related data.

   2 - Operation info. Fairly similar to the above, but on a smaller scale.
   Before/after maths operations, standard function calls etc.

   3 - High-level status output. This includes printing out contents of
   important informatic variables.

   4 - Med-level status output. F.ex. variable content from important variables
   which are maybe not so generally informatic as in level 3, but may still
   be important for debugging.

   5 - Low-level status output. F.ex. very low-level variable output like I/O
   inputed characters, very basic mathematical output. Often this level of
   output will not be needed unless some really obscure bugs occur. 
   
   Each level will also have a slightly different output to emphasize the
   level. This can be seen quite clearly from the macros below.
*/

#define D1PRINTF(x) { if(db_level>=1) {printDbgPrefix(1);eprintf x; }}
#define D2PRINTF(x) { if(db_level>=2) {printDbgPrefix(2);eprintf x; }}
#define D3PRINTF(x) { if(db_level>=3) {printDbgPrefix(3);eprintf x; }}
#define D4PRINTF(x) { if(db_level>=4) {printDbgPrefix(4);eprintf x; }}
#define D5PRINTF(x) { if(db_level>=5) {printDbgPrefix(5);eprintf x; }}

/* Print, whatever the debug level */
#define DPRINTF(x) { eprintf_nnl x; }

#else


/* Set dummy macros so code isn't broken when no DEBUG is not defined */

#define D1PRINTF(x)
#define D2PRINTF(x)
#define D3PRINTF(x)
#define D4PRINTF(x)
#define D5PRINTF(x)

#define DPRINTF(x)
#endif


#define DB_OUTMAX 4095  /* Maximum length of text one can print out with
			   one DnPRINTF call */



/* External variables */

extern int db_level;
extern FILE *db_fp;


/* Function prototypes */

int eprintf(char *fmt, ...);
int eprintf_nnl(char *fmt, ...);
void printDbgPrefix(int level);
void setDebugFile(FILE *fp);

#endif  /* End of #ifdef DEBUG_H */


