/** Filter management.
 *
 * Authors: Kristoffer Lawson, setok@kasvua.org
 * $Header: /home/sleber/cvs/liboil/filters.c,v 1.1.1.1 2003/01/06 02:09:05 sleber Exp $
 */

#include <SDL/SDL.h>
#include <tcl.h>
#include <stdlib.h>
#include "filters.h"


/* Static variables. */

/* Contains filters. Key: name of filter. Value: Filter structure. */
static Tcl_HashTable Filters;


/** Initialise filter library. Must be called before use. */

void Oil_InitFilterLib(void) {
  Tcl_InitHashTable(&Filters, TCL_STRING_KEYS);
}


/** Adds filter to pool. 
 *
 * @param  name  Identifier of filter.
 * @param  createFunc  Function used to create an instance of filter.
                       'originalSurface' is the original SDL surface of the
		       Oil object. Used mainly to get the format and size.

		       'options'  Hash table with filter-specific options
		       as key-value pairs. NULL if no options.

		       'filterData' is set by the creation functin to point
		       to filter-specific data which should be passed
		       to the rendering function on calling.

		       Returns surface that can be used by filter when 
		       rendering results, or NULL if an error occured.
 * @param  renderFunc  Function called when the filter's effect should be
                       rendered.
		       'originalSurface' -- the original SDL surface to read
		       from for generating effect.
		       'renderSurface' -- SDL surface to render effect to
		       'x' -- X offset for render.
		       'y' -- Y offset for render.
		       'filterData' -- filter-specific data, as specified
		       in 'createFunc'.

		       Note that renderSurface will not be automatically
		       cleared by the engine. If it needs clearing, the
		       filter should take care of this itself (otherwise
		       it will contain the previous frame -- which can
		       be used).
*/

void Oil_AddFilter(char *name, 
		   SDL_Surface *createFunc(SDL_Surface *originalSurface, 
					   Tcl_HashTable *options,
					   void **filterData),
		   void renderFunc(SDL_Surface *originalSurface, 
				   SDL_Surface *renderSurface,
				   int offsetX, int offsetY,
				   void *filterData)) {
  Oil_Filter *filterData;
  int newEntry;
  Tcl_HashEntry *hashEntry;

  filterData = malloc(sizeof(Oil_Filter));
  filterData->createFunc = createFunc;
  filterData->renderFunc = renderFunc;

  hashEntry=Tcl_CreateHashEntry(&Filters, name, &newEntry);
  Tcl_SetHashValue(hashEntry, filterData);

  return;
}


/** Finds the Oil_Filter structure for the filter by 'name' or NULL if
    no such filter exists. */

Oil_Filter *Oil_FindFilter(char *name) {
  Tcl_HashEntry *hashEntry;

  hashEntry=Tcl_FindHashEntry(&Filters, name);
  if (hashEntry == NULL) {
    return NULL;
  } else {
    return Tcl_GetHashValue(hashEntry);
  }
}



  

  



