/** 
 * Feedback with images being slowly distorted. Based on the effect used
 * for Sleber's CAFe 2002 demo.
 *
 * Authors: Kristoffer Lawson, setok@fishpool.com
 *
 * $Header: /home/sleber/cvs/liboil/filters/feedbackFilter.c,v 1.2 2003/02/17 01:22:50 setok Exp $
 */


#include <SDL/SDL.h>
#include <tcl.h>
#include "filters.h"
#include "../oil.h"
#include "../debug.h"


/* Macros. */

/* Size of random data buffer. */
#define PIXELRND_BUF 1024


/* Types. */

typedef struct {
  short x;
  short y;
} PixelLocation;

typedef struct {
  /* Pre-calced pixel offsets for surface. */
  int *pixelSrcsX;
  int *pixelSrcsY;

  /* Temporary surface. "Current" result is copied to it and red through
     for feedback. This way the rendering doesn't interfere with its
     source -- the previous frame. */  
  SDL_Surface *tmpSurface;
} FeedbackData;


/* Variables. */

/* Contain pre-calculated random values for offseting to avoid call to
   rand(). */
int PixelRnd[PIXELRND_BUF];
int PixelRndY[PIXELRND_BUF];

/* Index of random value to use for next draw. */
int RndIndex;


/* Function prototypes. */

SDL_Surface *feedbackFilterCreate(SDL_Surface *originalSurface, 
			       Tcl_HashTable *options,
			       void **filterData);
static void feedbackFilterRender(SDL_Surface *originalSurface, 
			      SDL_Surface *renderSurface,
			      int offsetX, int offsetY,
			      void *filterData);
static Uint32 getpixel16(SDL_Surface *surface, int x, int y);
static void putpixel16(SDL_Surface *surface, int x, int y, Uint32 colour);


static Uint32 getpixel16(SDL_Surface *surface, int x, int y) {
  Uint8 *p;
  
  int bpp = surface->format->BytesPerPixel;
  p = (Uint8 *)surface->pixels + y * surface->pitch + x * bpp;

  return *(Uint16 *)p;   
}


static void putpixel16(SDL_Surface *surface, int x, int y, Uint32 colour) {
  Uint16 *bufp;

  bufp = (Uint16 *)surface->pixels + y*surface->pitch/2 + x;
  *bufp = colour;
}


int feedbackFilterInit() {
  int i;

  Oil_AddFilter("feedback", feedbackFilterCreate, feedbackFilterRender);

  /* Read in random numbers for small amounts in pixel src. */
  for(i=0; i < PIXELRND_BUF; i++) {
    PixelRnd[i] = rand() % 3;
  }
  for(i=0; i < PIXELRND_BUF; i++) {
    PixelRndY[i] = rand() % 3;
  }
  RndIndex = 0;

  return TCL_OK;
}


SDL_Surface *feedbackFilterCreate(SDL_Surface *originalSurface, 
			       Tcl_HashTable *options,
			       void **filterData) {
  FeedbackData *feedbackData;
  int x, y;
  SDL_Surface *tmpSurface;
  int areaWidth = 400;
  int areaHeight = 400;

  feedbackData = malloc(sizeof(FeedbackData));
  feedbackData->tmpSurface = Oil_CreateDisplaySurface16(areaWidth, areaHeight);
  tmpSurface = feedbackData->tmpSurface;
  feedbackData->pixelSrcsX = malloc(tmpSurface->w * tmpSurface->h *
				    sizeof(int));
  feedbackData->pixelSrcsY = malloc(tmpSurface->w * tmpSurface->h *
				    sizeof(int));

  /* Pre-calc table for source locations. */
  for(y=0; y < tmpSurface->w; y++) {
    for(x=0; x < tmpSurface->h; x++) {
      *(feedbackData->pixelSrcsX + x + (y*tmpSurface->w)) = 
	((x - (tmpSurface->w / 2)) * 0.9) + (tmpSurface->w / 2);
      *(feedbackData->pixelSrcsY + x + (y * tmpSurface->h)) = 
	((y - (tmpSurface->h / 2)) * 0.9) + (tmpSurface->h / 2);
    }
  }

  *filterData = feedbackData;

  D1PRINTF(("feedbackFilterCreate: return"));
  return Oil_CreateDisplaySurface16(areaWidth, areaHeight);
}


static void feedbackFilterRender(SDL_Surface *originalSurface, 
		       SDL_Surface *renderSurface, int offsetX, int offsetY,
		       void *filterData) {
  SDL_Rect dest;
  int x, y;
  FeedbackData *feedbackData = filterData;
  SDL_Surface *tmpSurface = feedbackData->tmpSurface;
  Uint32 pixel;
  
  /* :NOTE: This could probably be built so that only the lower half is
     blitted (as the top shouldn't have the interference problem), but 
     whether this would benefit much is unsure. */
  SDL_BlitSurface(renderSurface, NULL, tmpSurface, NULL);

  SDL_LockSurface(renderSurface);
  SDL_LockSurface(tmpSurface);
  for(y=0; y < tmpSurface->h; y++) {
    for(x=0; x < tmpSurface->w; x++) {
      RndIndex++;
      RndIndex &= PIXELRND_BUF-1;
	 
      pixel=getpixel16(tmpSurface, 
		       *(feedbackData->pixelSrcsX + x + (y * tmpSurface->w)) ^ 
		          PixelRnd[RndIndex], 
		       *(feedbackData->pixelSrcsY + x + (y * tmpSurface->w)) ^ 
		          PixelRndY[RndIndex]);
      putpixel16(renderSurface, x, y, pixel);
    }
  }
  SDL_UnlockSurface(renderSurface);
  SDL_UnlockSurface(tmpSurface);

  dest.x = (renderSurface->w / 2) - (originalSurface->w / 2);
  dest.y = (renderSurface->h / 2) - (originalSurface->h / 2);
  SDL_BlitSurface(originalSurface, NULL, renderSurface, &dest);
    
  return;
}
