/** Shell for liboil: a high-level library for controlling graphical elements 
    and FX on screen in the form of objects. 
    
    Objects are specified as a chain which are displayed in order. Each
    object can have individual filters attached, which can also be 
    chained to create a wide range of FX. Each object has an invidual 
    position and can be scaled, animated and moved. All of this happens
    with just a few Tcl commands making the use of this library as easy
    as humanly possible.

    Authors: Kristoffer Lawson, setok@fishpool.com

    $Header: /home/sleber/cvs/liboil/oilsh.c,v 1.3 2003/02/17 02:54:52 setok Exp $
*/


#include <tcl.h>
#include <SDL/SDL.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "debug.h"
#include "oil.h"
#include "filters.h"
// #include "blurFilter.h"
#include "filters/boxesFilter.h"
#include "colorCycleFilter.h"
#include "colorWobbleFilter.h"
#include "compColorFilter.h"
#include "filters/feedbackFilter.h"
#include "flashFilter.h"
#include "filters/kaleido.h"
#include "noiseFilter.h"
#include "wobbleFilter.h"
#include "wobble2Filter.h"
#include "wobble3Filter.h"
#include "wobble4Filter.h"
#include "wobble5Filter.h"
#include "xFlagFilter.h"
#include "xScrollFilter.h"
#include "xSplitFilter.h"
#include "xSplit2Filter.h"
#include "xStretchFilter.h"
#include "yFlagFilter.h"
#include "yScrollFilter.h"
#include "ySplitFilter.h"
#include "ySplit2Filter.h"
#include "yStretchFilter.h"

/* Types. */

/* Information for a created filter connected to an object. */
typedef struct FilterInstance {
  /* Contains data of filter like functions to call. */
  Oil_Filter *classData;

  /* Filter-specific data. */
  void *filterData;

  /* Surface filter renders to. */
  SDL_Surface *surface;

  /* Next filter in chain. NULL if none. */
  struct FilterInstance *next;
} FilterInstance;

/* Internal information for an Oil object. */
typedef struct OilOb {
  int root;  /* 1 if root object, 0 if not. */
  Tcl_Interp *interp;  /* Interpreter used in connection to this object. */

  int activated;  /* Is object active (= is rendered, animated and 
		     moved). */
  Uint8 alpha;  /* Object alpha value. */
  int x;  /* X co-ord of object. */
  int y;  /* Y co-ord. */
  int destroyed;  /* User requested destruction of ob. */
  char *name;  /* Identifying name of object. */
  SDL_Surface *surface;  /* Contains "original" data of object, which has
			    not yet been changed by filters or anything 
			    else. */
  struct OilOb *parent;  /* Object to render before this one. */
  struct OilOb *child;  /* Object to render after this one. */
  FilterInstance *filterHead;  /* First filter to be used to process 
				  object. */
  FilterInstance *filterTail;  /* Last filter to be processed for object. */

  /* Animation. */
  int animLength;
  SDL_Surface **frames;  /* Array of surfaces containing each frame. */
  int loop;  /* Do we loop the animation? */
  Uint32 clicksPerFrame;  /* Amount of milliseconds per frame. */
  int currentFrame;
  Uint32 lastFrameTime;  /* Time, in milliseconds, when last frame was 
			    shown. */
  int animSpeed;  /* Speed of animation, in frames per minute. */

  /* Movement */
  int toX;  /* X co-ordinate to move object to. */
  int toY;  /* Y co-ordinate to move object to. */
  int fromX;  /* Original X co-ordinate for move operation. */
  int fromY; /* Original Y co-ordinate for move. */
  int distanceX;  /* toX - fromX */
  int distanceY;  /* toY - fromY */
  unsigned long moveTime;
  int speed;  /* Speed at which to move in pixels/second. */
  Uint32 moveStartTime;
  //  Tcl_Time moveStartTime;
  Tcl_Obj *moveCallback;  /* Script to evaluate when movement is complete. */

  /* Scale to size. -1 if we do not want to scale. */
  int scaleWidth;  
  int scaleHeight;
} OilOb;


/* Macros. */

/* Prefix to use for string representation of Oil objects in Tcl. */
#define OILOB_STRPREFIX "oilOb"
/* Maximum length of a string capable of holding an integer. */
#define MAXNUMSTR 32
/* Name of root object. */
#define ROOT_NAME "root"


/* Function prototypes. */

static void animateOb(OilOb *oilData);
static void copyMovementInfo(OilOb *srcOb, OilOb *destOb);
static int demoEvent(SDL_Event *event);
static int draw();
static void drawScaledOb(SDL_Surface *renderTo, OilOb *ob);
static void filterInstanceFreeChain(FilterInstance *filterInstance);
static FilterInstance *filterInstanceNew(Oil_Filter *class, 
					 SDL_Surface *renderSurface,
					 void *filterData);
static void freeAnimFrames(OilOb *ob);
static SDL_Surface *getRenderingSurface(OilOb *ob);
void mainLoop(void);
static int moveOb(OilOb *oilData);
int Tcl_AppInit(Tcl_Interp *interp);
static int initFilters(Tcl_Interp *interp);

/* Functions for handling Oil objects. */
void oilOb_Dup(Tcl_Obj *src, Tcl_Obj *dst);
void oilOb_Free(Tcl_Obj *ob);
static Tcl_Obj *oilOb_New(OilOb *parent, char *name, Tcl_Interp *interp, 
			  SDL_Surface *surface, 
			  int x, int y);
int oilOb_Set(Tcl_Interp *interp, Tcl_Obj *obj);
void oilOb_UpdateString(Tcl_Obj *obj);
static OilOb *getOilData(Tcl_Obj *ob);

/* New Tcl commands. */
static int activateCmd(ClientData clientData, Tcl_Interp *interp, int objc,
		       Tcl_Obj *const objv[]);
static int clearObCmd(ClientData clientData, Tcl_Interp *interp, int objc,
	    Tcl_Obj *const objv[]);
static int copyFramesCmd(ClientData clientData, Tcl_Interp *interp, int objc,
			 Tcl_Obj *const objv[]);
int createOilOb(ClientData clientData, Tcl_Interp *interp, int objc,
		Tcl_Obj *const objv[]);
static int deactivateCmd(ClientData clientData, Tcl_Interp *interp, int objc,
			 Tcl_Obj *const objv[]);
static int destroyObCmd(ClientData clientData, Tcl_Interp *interp, int objc,
			 Tcl_Obj *const objv[]);
static int drawCmd(ClientData clientData, Tcl_Interp *interp, int objc,
		   Tcl_Obj *const objv[]);
static int filterCmd(ClientData clientData, Tcl_Interp *interp, int objc,
		     Tcl_Obj *const objv[]);
static int getHeightCmd(ClientData clientData, Tcl_Interp *interp, int objc,
			Tcl_Obj *const objv[]);
static int getPosCmd(ClientData clientData, Tcl_Interp *interp, int objc,
		     Tcl_Obj *const objv[]);
int listObs(ClientData clientData, Tcl_Interp *interp, int objc,
	    Tcl_Obj *const objv[]);
static int loadBitmap(ClientData clientData, Tcl_Interp *interp, int objc,
		      Tcl_Obj *const objv[]);
static int loadAnimFramesCmd(ClientData clientData, Tcl_Interp *interp, 
			     int objc,
			     Tcl_Obj *const objv[]);
static int makeRootCmd(ClientData clientData, Tcl_Interp *interp, int objc,
		    Tcl_Obj *const objv[]);
static int moveCmd(ClientData clientData, Tcl_Interp *interp, int objc,
		   Tcl_Obj *const objv[]);
static int rmFiltersCmd(ClientData clientData, Tcl_Interp *interp, int objc,
		       Tcl_Obj *const objv[]);
static int scaleCmd(ClientData clientData, Tcl_Interp *interp, int objc,
		       Tcl_Obj *const objv[]);
static int setAlphaCmd(ClientData clientData, Tcl_Interp *interp, int objc,
		       Tcl_Obj *const objv[]);
static int setAnimSpeedCmd(ClientData clientData, Tcl_Interp *interp, int objc,
		       Tcl_Obj *const objv[]);
static int setColourKeyCmd(ClientData clientData, Tcl_Interp *interp, int objc,
			   Tcl_Obj *const objv[]);
static int setFrameCmd(ClientData clientData, Tcl_Interp *interp, int objc,
		       Tcl_Obj *const objv[]);
static int setMoveSpeedCmd(ClientData clientData, Tcl_Interp *interp, int objc,
			   Tcl_Obj *const objv[]);
static int setParentCmd(ClientData clientData, Tcl_Interp *interp, int objc,
			Tcl_Obj *const objv[]);
static int setPosCmd(ClientData clientData, Tcl_Interp *interp, int objc,
		     Tcl_Obj *const objv[]);
int setVideo(ClientData clientData, Tcl_Interp *interp, int objc, 
	     Tcl_Obj *const objv[]);


/* Global variables. */

/* Holds details of a OilObType object description. */
Tcl_ObjType OilObType;

/* Amount of created OilObTypes (not that this does not decrease when
   they are deleted). */
static int CreatedOilObs=0;

/* Keeps mapping of strings to Oil object internal structure. */
static Tcl_HashTable Str2OilMap;

/* Framebuffer surface. */
static SDL_Surface *FBSurface=NULL;

/* Root object data. */
static OilOb *RootObData;

/* Last time rendering speed was checked. */
static Uint32 LastCheckTime=0;

/* Colour to fill root with. */
static Uint32 ClearColour;


int main(int argc, char *argv[]) {
  Tcl_SetMainLoop(mainLoop);
  Tcl_Main(argc, argv, Tcl_AppInit);

  return 0;
}



/* Initialise interpreter. */

int Tcl_AppInit(Tcl_Interp *interp) {
  int rc;

  setDebugFile(stdout);
  setbuf(stdout, NULL);
  SDL_Init(SDL_INIT_VIDEO|SDL_INIT_NOPARACHUTE);
  atexit(SDL_Quit);

  Tcl_Init(interp);
  Tcl_CreateObjCommand(interp, "activate", activateCmd, NULL, NULL);
  Tcl_CreateObjCommand(interp, "clearOb", clearObCmd, NULL, NULL);
  Tcl_CreateObjCommand(interp, "copyFrames", copyFramesCmd, NULL, NULL);
  Tcl_CreateObjCommand(interp, "createOb", createOilOb, NULL, NULL);
  Tcl_CreateObjCommand(interp, "deactivate", deactivateCmd, NULL, NULL);
  Tcl_CreateObjCommand(interp, "destroyOb", destroyObCmd, NULL, NULL);
  Tcl_CreateObjCommand(interp, "draw", drawCmd, NULL, NULL);
  Tcl_CreateObjCommand(interp, "getHeight", getHeightCmd, NULL, NULL);
  Tcl_CreateObjCommand(interp, "getPos", getPosCmd, NULL, NULL);
  Tcl_CreateObjCommand(interp, "makeRoot", makeRootCmd, NULL, NULL);
  Tcl_CreateObjCommand(interp, "rmFilters", rmFiltersCmd, NULL, NULL);  
  Tcl_CreateObjCommand(interp, "scale", scaleCmd, NULL, NULL);  
  Tcl_CreateObjCommand(interp, "setAlpha", setAlphaCmd, NULL, NULL);  
  Tcl_CreateObjCommand(interp, "setAnimSpeed", setAnimSpeedCmd, NULL, NULL);  
  Tcl_CreateObjCommand(interp, "setColourKey", setColourKeyCmd, NULL, 
		       NULL);  
  Tcl_CreateObjCommand(interp, "setFrame", setFrameCmd, NULL, NULL); 
  Tcl_CreateObjCommand(interp, "setMoveSpeed", setMoveSpeedCmd, NULL, NULL); 
  Tcl_CreateObjCommand(interp, "setParent", setParentCmd, NULL, NULL); 
  Tcl_CreateObjCommand(interp, "setVideo", setVideo, NULL, NULL); 
  Tcl_CreateObjCommand(interp, "loadAnimFrames", loadAnimFramesCmd, NULL, 
		       NULL);
  Tcl_CreateObjCommand(interp, "loadBitmap", loadBitmap, NULL, NULL); 
  Tcl_CreateObjCommand(interp, "listObs", listObs, NULL, NULL); 
  Tcl_CreateObjCommand(interp, "filter", filterCmd, NULL, NULL);
  Tcl_CreateObjCommand(interp, "setPos", setPosCmd, NULL, NULL);
  Tcl_CreateObjCommand(interp, "move", moveCmd, NULL, NULL);

  OilObType.name = "oilObType";
  OilObType.freeIntRepProc = oilOb_Free;
  OilObType.dupIntRepProc = oilOb_Dup;
  OilObType.updateStringProc = oilOb_UpdateString;
  OilObType.setFromAnyProc = oilOb_Set;      
  Tcl_RegisterObjType(&OilObType);
  Tcl_InitHashTable(&Str2OilMap, TCL_STRING_KEYS);

  db_level = 5;
  
  if((rc=initFilters(interp)) != TCL_OK) {
    return rc;
  } 

  return Tcl_PkgProvide(interp, "org.sleberEid.oil", "0.1.1");
}


static int initFilters(Tcl_Interp *interp) {
  Oil_InitFilterLib();

  //  blurFilterInit();
  boxesFilterInit();
  colorCycleFilterInit();
  colorWobbleFilterInit();
  compColorFilterInit();
  feedbackFilterInit();
  flashFilterInit();
  kaleidoInit();
  noiseFilterInit();
  wobbleFilterInit();
  wobble2FilterInit();
  wobble3FilterInit();
  wobble4FilterInit();
  wobble5FilterInit();
  xFlagFilterInit();
  xScrollFilterInit();
  xSplitFilterInit();
  xSplit2FilterInit();
  xStretchFilterInit();
  yFlagFilterInit();
  yScrollFilterInit();
  ySplitFilterInit();
  ySplit2FilterInit();
  yStretchFilterInit();

  return TCL_OK;
}


static void putpixel16(SDL_Surface *surface, int x, int y, Uint32 colour) {
  Uint16 *bufp;

  bufp = (Uint16 *)surface->pixels + y*surface->pitch/2 + x;
  *bufp = colour;
}


static void putpixel32(SDL_Surface *surface, int x, int y, Uint32 pixel) {
  Uint8 *p;

  int bpp = surface->format->BytesPerPixel;
  p = (Uint8 *)surface->pixels + y * surface->pitch + x * bpp;
  
  *(Uint32 *)p = pixel;
}


/** Create an Oil object. 
 *
 * @param  parent  Parent object. The created object will be drawn after the
 *                 parent. If the parent had any other child it will be made
 *                 the child of the newly created object. If NULL, object
 *                 has no parent.
 * @param  name  Name of object. This must point to a malloced area.
 * @param  x  X co-ordinate of object at creation. Relative to screen.
 * @param  y  Y co-ordinate of object at creation. Relative to screen.
 * @parem  w  Initial width of object.
 * @param  h  Initial height of object.
 *
 * @return  New Tcl_Obj representing created Oil object.
 */

static Tcl_Obj *oilOb_New(OilOb *parent, char *name, Tcl_Interp *interp, 
			  SDL_Surface *surface, 
			  int x, int y) {
  Tcl_HashEntry *hashEntry;
  int newEntry;
  Tcl_Obj *ob = (Tcl_Obj *) Tcl_Alloc(sizeof(Tcl_Obj));
  OilOb *oilStruct = (OilOb *) Tcl_Alloc(sizeof(OilOb));
  
  D1PRINTF(("oilOb_new called"));

  ob->refCount=0;  
  ob->bytes=Tcl_Alloc(strlen(name)+1);
  strcpy(ob->bytes, name);
  ob->length = strlen(name);
  ob->typePtr = &OilObType;

  oilStruct->root = 0;  /* By default we're not a root object. Change
			   this if we are. */
  oilStruct->interp = interp;
  oilStruct->surface = surface;
  oilStruct->parent = parent;
  oilStruct->activated = 1;

  oilStruct->alpha = SDL_ALPHA_OPAQUE;
  oilStruct->x = x;
  oilStruct->y = y;
  oilStruct->destroyed = 0;
  oilStruct->name = name;
  oilStruct->filterHead = NULL;
  oilStruct->filterTail = NULL;

  oilStruct->toX = 0;
  oilStruct->toY = 0;
  oilStruct->speed = 0;
  oilStruct->moveCallback = NULL;

  oilStruct->animLength = 0;
  oilStruct->animSpeed = 0;
  oilStruct->frames = NULL;

  oilStruct->scaleWidth = -1;
  oilStruct->scaleHeight = -1;

  //  oilStruct->moveStartTime = 0;

  if (parent != NULL) {
    /* Update parent and object's child, if necessary. The object is inserted
       between the parent and any existing child.
    */
    oilStruct->child = parent->child;
    parent->child = oilStruct;    
    if(oilStruct->child != NULL) {
      /* Update any existing child to point to parent. */
      oilStruct->child->parent = oilStruct;
    }
  } else {
    oilStruct->child = NULL;
  }

  ob->internalRep.otherValuePtr = oilStruct;

  /* Register internal structure in hash. We might need to look it up again
     from a string. */
  D3PRINTF(("%s", ob->bytes));
  hashEntry=Tcl_CreateHashEntry(&Str2OilMap, ob->bytes, &newEntry);
  Tcl_SetHashValue(hashEntry, oilStruct);

  D1PRINTF(("oilOb_New: return"));
  return ob;
}


/** Free an Oil object. */

void oilOb_Free(Tcl_Obj *ob) {
  D1PRINTF(("oilOb_Free called"));

  return;
}


/** Duplicate a Oil object. */

void oilOb_Dup(Tcl_Obj *src, Tcl_Obj *dst) {
  D1PRINTF(("oilOb_Dup called"));
  dst->internalRep.otherValuePtr = src->internalRep.otherValuePtr;

  return;
}
  
 
/** Generate valid Oil object from string representation. */

int oilOb_Set(Tcl_Interp *interp, Tcl_Obj *obj) {
  Tcl_HashEntry *hashEntry;
  OilOb *internalRep;
  Tcl_Obj *resObj;
  Tcl_ObjType *oldType;
  char *obString;

  D1PRINTF(("oilOb_Set called"));

  D2PRINTF(("oilOb_Set: get string from object"));
  obString=Tcl_GetString(obj);
  D3PRINTF(("oilOb_Set: obString: %s", obString));
  D2PRINTF(("oilOb_Set: search for ob in hash table"));

  hashEntry=Tcl_FindHashEntry(&Str2OilMap, obString);
  if(hashEntry != NULL) {
    D2PRINTF(("oilOb_Set: found object from hash table"));
    /* Found key in string->oil hash table. */
    oldType = obj->typePtr;
    obj->typePtr = &OilObType;

    if((oldType != NULL) && (oldType->freeIntRepProc != NULL)) {
      D4PRINTF(("oilOb_Set: oldType->name: %s", oldType->name));
      D2PRINTF(("oilOb_Set: free old representation"));
      /* Free previous representation. */
      oldType->freeIntRepProc(obj);
    }

    D2PRINTF(("oilOb_Set: get hash value"));
    internalRep=Tcl_GetHashValue(hashEntry);
    obj->internalRep.otherValuePtr=internalRep;

    D1PRINTF(("oilOb_Set returns"));
    return TCL_OK;
  } else {
    D2PRINTF(("oilOb_Set: could not find object from hash table"));
    if(interp) {
      resObj = Tcl_NewStringObj("Could not find Oil object: ", -1);
      Tcl_AppendToObj(resObj, obj->bytes, obj->length);
      Tcl_SetObjResult(interp, resObj);
    }

    D1PRINTF(("oilOb_Set returns"));
    return TCL_ERROR;
  }
}


/** Generate valid string representation of Oil object. */

void oilOb_UpdateString(Tcl_Obj *obj) {
  OilOb *internalRep=obj->internalRep.otherValuePtr;

  D1PRINTF(("oilOb_UpdateString called"));
  obj->bytes = internalRep->name;
  obj->length = strlen(internalRep->name);

  return;
}


/** 
 * Frees a filter instance.
 */

static void filterInstanceFreeChain(FilterInstance *filterInstance) {
  D1PRINTF(("filterInstanceFreeChain called"));

  if(filterInstance == NULL) {
    return;
  }

  filterInstanceFreeChain(filterInstance->next);

  if(filterInstance->filterData) {
    /* Bug here: filters should free their filterData as it may itself
       contain dynamically allocated data. */
    free(filterInstance->filterData);
  }
  SDL_FreeSurface(filterInstance->surface);

  return;
}


/**
 * Copies movement information of 'srcOb' to 'destOb'.
 */

static void copyMovementInfo(OilOb *srcOb, OilOb *destOb) {
  destOb->toX = srcOb->toX;
  destOb->toY = srcOb->toY;
  destOb->fromX = srcOb->fromX;
  destOb->fromY = srcOb->fromY;
  destOb->distanceX = srcOb->distanceX;
  destOb->distanceY = srcOb->distanceY;
  destOb->moveTime = srcOb->moveTime;
  destOb->speed = srcOb->speed;
  destOb->moveStartTime = srcOb->moveStartTime;
  destOb->moveCallback = srcOb->moveCallback;

  return;
}


/**
 * Creates and initialises new FilterInstance structure. 
 */

static FilterInstance *filterInstanceNew(Oil_Filter *class, 
					 SDL_Surface *renderSurface,
					 void *filterData) {
  FilterInstance *new;

  new = malloc(sizeof(FilterInstance));
  new->classData = class;
  new->surface = renderSurface;
  new->filterData = filterData;
  new->next = NULL;

  return new;
}


/**
 * Frees animation frames from 'ob'.
 */

static void freeAnimFrames(OilOb *ob) {
  int i;

  for(i=0; i < ob->animLength; i++) {
    SDL_FreeSurface(ob->frames[i]);
  }

  return;
}


/** 
 * Copies filters from 'srcOb' to 'destOb' (makes new instances of the
 * filters. 
 */

static void copyFilters(OilOb *srcOb, OilOb *destOb) {
  FilterInstance *srcFilterInstance, *dstFilterInstance; 
  SDL_Surface *filterSurface;
  void *filterData;

  filterInstanceFreeChain(destOb->filterHead);

  if(srcOb->filterHead != NULL) {
    /* Filters exist. */

    /* Create first filter. */
    filterData = NULL;
    srcFilterInstance = srcOb->filterHead;
    filterSurface = 
      srcFilterInstance->classData->createFunc(destOb->surface,
					       NULL, 
					       &filterData);
    destOb->filterHead = 
      filterInstanceNew(srcFilterInstance->classData, filterSurface, 
			filterData);

    /* Rest of the filters. */
    srcFilterInstance = srcOb->filterHead->next;
    dstFilterInstance = destOb->filterHead;
    while(srcFilterInstance != NULL) {
      filterData = NULL;
      filterSurface = srcFilterInstance->classData->createFunc(filterSurface, 
							       NULL, 
							       &filterData);
      dstFilterInstance->next = 
	filterInstanceNew(srcFilterInstance->classData, filterSurface, 
			  filterData);

      srcFilterInstance = srcFilterInstance->next;
      dstFilterInstance = dstFilterInstance->next;
    }

    destOb->filterTail = dstFilterInstance;
  }

  return;
}


/** 
 * Converts object into the root object. All aspects of the object are
 * copied to the root object and after this command the object will no
 * longer exist.
 *
 * The children of the object become children of the object's parent.
 * If the object has no parent, the children become orphans.
 *
 * Usage: makeRoot <object>
 */

int makeRootCmd(ClientData clientData, Tcl_Interp *interp, int objc,
		Tcl_Obj *const objv[]) {
  OilOb *oilOb;
  Tcl_HashEntry *hashEntry;
  SDL_Surface *newRootSurface;

  if(objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "object");
    return TCL_ERROR;
  }
   
  if(Tcl_ConvertToType(interp, objv[1], &OilObType) == TCL_ERROR) {
    return TCL_ERROR;
  }
  oilOb = objv[1]->internalRep.otherValuePtr;

  if(oilOb == RootObData) {
    /* Passed object already was root object! Do nothing. */
    return TCL_OK;
  }

  /* Move surface. */
  if(RootObData->surface != FBSurface) {
    /* Previous surface was not the framebuffer surface, so free it first. */
    SDL_FreeSurface(RootObData->surface);
  }
  
  /* Make sure surface is right size. */
  if((newRootSurface = Oil_CreateDisplaySurface16(FBSurface->w, FBSurface->h))
     == NULL) {
    Tcl_SetObjResult(interp, Tcl_NewStringObj(SDL_GetError(), -1));
    return TCL_ERROR;
  }    
  SDL_SetAlpha(oilOb->surface, 0, 0);
  SDL_BlitSurface(oilOb->surface, NULL, newRootSurface, NULL);
  SDL_FreeSurface(oilOb->surface);
  RootObData->surface = newRootSurface;

  /* Move frames. */
  freeAnimFrames(RootObData);
  RootObData->frames = oilOb->frames;    
  RootObData->animLength = oilOb->animLength;
  RootObData->loop = oilOb->loop;
  RootObData->clicksPerFrame = oilOb->clicksPerFrame;
  RootObData->currentFrame = oilOb->currentFrame;
  RootObData->lastFrameTime = oilOb->lastFrameTime;
  RootObData->animSpeed = oilOb->animSpeed;

  /* Copy filters. */
  copyFilters(oilOb, RootObData);
  filterInstanceFreeChain(oilOb->filterHead);

  /* Copy movement info. */
  copyMovementInfo(oilOb, RootObData);

  /* Copy attributes. */
  RootObData->alpha = oilOb->alpha;
  RootObData->scaleWidth = oilOb->scaleWidth;
  RootObData->scaleHeight = oilOb->scaleHeight;

  /* Destroy object. */

  hashEntry=Tcl_FindHashEntry(&Str2OilMap, oilOb->name);
  if(hashEntry != NULL) {
    Tcl_DeleteHashEntry(hashEntry);
  }

  free(oilOb->name);
  if(oilOb->moveCallback != NULL) {
    Tcl_DecrRefCount(oilOb->moveCallback);
  }
  if(oilOb->parent != NULL) {
    oilOb->parent->child = oilOb->child;
  }

  Tcl_Free((char *) oilOb);
  objv[1]->internalRep.otherValuePtr=NULL;
  objv[1]->typePtr=NULL;

  return TCL_OK;
}


/** 
 * Returns a list containing the X and Y position of the object.
 *
 * Usage: getPos <object> 
 */

int getPosCmd(ClientData clientData, Tcl_Interp *interp, int objc,
		 Tcl_Obj *const objv[]) {
  OilOb *oilData;
  Tcl_Obj *list;

  if(objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "object");
    return TCL_ERROR;
  }
   
  if(Tcl_ConvertToType(interp, objv[1], &OilObType) == TCL_ERROR) {
    return TCL_ERROR;
  }
  oilData = objv[1]->internalRep.otherValuePtr;

  list = Tcl_NewListObj(0, NULL);
  
  /* X position. */
  if (Tcl_ListObjAppendElement(interp, list, Tcl_NewIntObj(oilData->x)) 
      == TCL_ERROR) {
    return TCL_ERROR;
  }

  /* Y position. */
  if(Tcl_ListObjAppendElement(interp, list, Tcl_NewIntObj(oilData->y)) 
     == TCL_ERROR) {
    return TCL_ERROR;
  }
  
  Tcl_SetObjResult(interp, list);
  return TCL_OK;
}


/** 
 * Returns the height object is currently scaled to.
 *
 * Usage: getHeight <object> 
 */

int getHeightCmd(ClientData clientData, Tcl_Interp *interp, int objc,
		 Tcl_Obj *const objv[]) {
  OilOb *oilData;

  if(objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "object");
    return TCL_ERROR;
  }
   
  if(Tcl_ConvertToType(interp, objv[1], &OilObType) == TCL_ERROR) {
    return TCL_ERROR;
  }
  oilData = objv[1]->internalRep.otherValuePtr;

  if(oilData->scaleHeight != -1) {
    Tcl_SetObjResult(interp, Tcl_NewIntObj(oilData->scaleHeight));
  } else {
    Tcl_SetObjResult(interp, Tcl_NewIntObj(getRenderingSurface(oilData)->h));
  }

  return TCL_OK;
}

  
/** 
 * Clear object surface with given colour.
 *
 * Usage: clearOb <object> <r> <g> <b>
 * 
 * 'r', 'g', 'b'  Colour to clear surface with.
 */

int clearObCmd(ClientData clientData, Tcl_Interp *interp, int objc,
	       Tcl_Obj *const objv[]) {
  OilOb *oilData;
  int r, g, b;
  Uint32 clearColour;

  if(objc < 5) {
    Tcl_WrongNumArgs(interp, 1, objv, "object");
    return TCL_ERROR;
  }
   
  if(Tcl_ConvertToType(interp, objv[1], &OilObType) == TCL_ERROR) {
    return TCL_ERROR;
  }
  oilData = objv[1]->internalRep.otherValuePtr;

  if(Tcl_GetIntFromObj(interp, objv[2], &r) == TCL_ERROR) {
    return TCL_ERROR;
  }
  if(Tcl_GetIntFromObj(interp, objv[3], &g) == TCL_ERROR) {
    return TCL_ERROR;
  }
  if(Tcl_GetIntFromObj(interp, objv[3], &b) == TCL_ERROR) {
    return TCL_ERROR;
  }

  if (oilData->root) {
    if(oilData->surface != FBSurface) {
      SDL_FreeSurface(oilData->surface);
      oilData->surface = FBSurface;
    }
    ClearColour = SDL_MapRGB(FBSurface->format, r, g, b);
  } else {
    clearColour=SDL_MapRGB(oilData->surface->format, r, g, b);
    SDL_FillRect(oilData->surface, NULL, clearColour);
  }
  
  return TCL_OK;
}
  

/** 
 * Removes all filters from object.
 *
 * Usage: rmFilters <object>
 */

int rmFiltersCmd(ClientData clientData, Tcl_Interp *interp, int objc,
		 Tcl_Obj *const objv[]) {
  OilOb *oilData;

  if(objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "object");
    return TCL_ERROR;
  }
   
  if(Tcl_ConvertToType(interp, objv[1], &OilObType) == TCL_ERROR) {
    return TCL_ERROR;
  }
  oilData = objv[1]->internalRep.otherValuePtr;

  if(oilData->filterHead != NULL) {
    if(oilData->filterTail->surface->flags & SDL_SRCCOLORKEY) {
      /* Colour key set for last filter's render surface: should set base 
	 surface accordingly. */
      /* :NOTE: This could be considered a bug: what if setting a colour key
	 is part of the filter's logic? Maybe the object itself should contain
	 the colour key information after all. */
      SDL_SetColorKey(oilData->surface, SDL_SRCCOLORKEY, 
		      oilData->filterTail->surface->format->colorkey);
    }
    filterInstanceFreeChain(oilData->filterHead);
  }

  oilData->filterHead = NULL;
  oilData->filterTail = NULL;

  return TCL_OK;
}


/** 
 * Destroys object.
 *
 * Usage: destroyOb <object>
 */

int destroyObCmd(ClientData clientData, Tcl_Interp *interp, int objc,
		 Tcl_Obj *const objv[]) {
  OilOb *oilData;
  Tcl_HashEntry *hashEntry;
  int i;

  if(objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "object");
    return TCL_ERROR;
  }
   
  if(Tcl_ConvertToType(interp, objv[1], &OilObType) == TCL_ERROR) {
    return TCL_ERROR;
  }
  oilData = objv[1]->internalRep.otherValuePtr;

  D3PRINTF(("destroyObCmd: oilData->name: %s", oilData->name));
  hashEntry=Tcl_FindHashEntry(&Str2OilMap, oilData->name);
  if(hashEntry != NULL) {
    D2PRINTF(("destroyObCmd: found object from hash table -- delete"));
    Tcl_DeleteHashEntry(hashEntry);
  }

  if(oilData->parent != NULL) {
    oilData->parent->child = oilData->child;
  }
  if(oilData->child != NULL) {
    oilData->child->parent = oilData->parent;
  }

  D2PRINTF(("destroyObCmd: free name"));
  free(oilData->name);
  D2PRINTF(("destroyObCmd: check callback script"));  
  if(oilData->moveCallback != NULL) {
    Tcl_DecrRefCount(oilData->moveCallback);
  }

  D1PRINTF(("destroyObCmd: check animation"));
  if(oilData->animLength != 0) {
    D2PRINTF(("destroyObCmd: free animation frames"));
    D3PRINTF(("destroyObCmd: animLength: %d", oilData->animLength));
    /* Free animation frames. */
    for (i=0; i < oilData->animLength; i++) {
      D4PRINTF(("destroyObCmd: i: %d", i));
      SDL_FreeSurface(oilData->frames[i]);
    }
    D2PRINTF(("destroyObCmd: free frame array"));
    free(oilData->frames);
  } else {
      SDL_FreeSurface(oilData->surface);
  }

  D1PRINTF(("destroyObCmd: check filters"));
  if(oilData->filterHead != NULL) {
    filterInstanceFreeChain(oilData->filterHead);
  }
  
  D1PRINTF(("destroyObCmd: free actual data"));
  Tcl_Free((char *) oilData);
  objv[1]->internalRep.otherValuePtr=NULL;
  objv[1]->typePtr=NULL;

  D1PRINTF(("destroyObCmd: return"));
  return TCL_OK;
}


/** 
 * Implements command to create Oil object. 
 *
 * Usage: createOb <parent>
 *
 * Returns identifier of object. Should probably be set to some variable
 * to avoid unnecessary destruction and re-creation.
 */

int createOilOb(ClientData clientData, Tcl_Interp *interp, int objc,
		Tcl_Obj *const objv[]) {
  OilOb *parentOb;
  char *name;
  SDL_Surface *surface;
  int width = 0, height = 0;

  D1PRINTF(("create called"));
  if(objc < 2) {
    /* No parent given. Create orphan. */
    parentOb = NULL;
  } else {   
    if(Tcl_ConvertToType(interp, objv[1], &OilObType) == TCL_ERROR) {
      return TCL_ERROR;
    }
    parentOb = objv[1]->internalRep.otherValuePtr;
  }

  if(objc > 2) {
    /* Size argument given. */
    if (objc != 4) {
      Tcl_WrongNumArgs(interp, 1, objv, "parent ?width height?");
      return TCL_ERROR;
    }

    if(Tcl_GetIntFromObj(interp, objv[2], &width) == TCL_ERROR) {
      return TCL_ERROR;
    }
    if(Tcl_GetIntFromObj(interp, objv[3], &height) == TCL_ERROR) {
      return TCL_ERROR;
    }
  }

  /* Create a name for the object. */
  name = malloc(strlen(OILOB_STRPREFIX) + MAXNUMSTR + 1);  
  strcpy(name, OILOB_STRPREFIX);
  sprintf(name+strlen(OILOB_STRPREFIX), "%d", CreatedOilObs);
  CreatedOilObs++;

  surface = Oil_CreateDisplaySurface16(width, height);
  if(surface == NULL) {
    /* Error when creating surface. See SDL error for info. */

    Tcl_SetObjResult(interp, Tcl_NewStringObj(SDL_GetError(), -1));
    return TCL_ERROR;
  }
  Tcl_SetObjResult(interp, oilOb_New(parentOb, name, interp, surface, 0, 
				     0));

  return TCL_OK;
}


/** Returns the internal Oil structure from the Tcl object 'ob'. */

static OilOb *getOilData(Tcl_Obj *ob) {
  return ob->internalRep.otherValuePtr;
}


/*
 *  Handles *event and returns false. If event == null, polls for one
 *  event and handles it. Returns true if there are pending events.
 */

static int demoEvent(SDL_Event *event) {
  SDL_Event polled_event;
  int pending = 0;

  if ( event == NULL ) {
    pending = SDL_PollEvent(&polled_event);
    event = &polled_event;
  }

  switch ( event->type ) {
  case SDL_KEYDOWN:
    if ( event->key.keysym.sym == SDLK_ESCAPE ) {
      exit(0);
    }
    break;
  case SDL_QUIT:
    exit(0);
    break;
  default:
    break;
  }

  return pending;
}


/** Draws objects on screen and handles SDL events. */

static int draw() {
  static int frameCount = 0;
  OilOb *obData;
  SDL_Rect destRect;
  SDL_Surface *obSurface;
  FilterInstance *filterStack;
  int flags;

  demoEvent(NULL);

  if(FBSurface == NULL) {
    /* Not yet initialised. */
    return 0;
  }

  /* Handle root object. */

  if (RootObData->surface != FBSurface) {
    /* Has alternative background. Fill with it instead of black and
       include filters, if any set. */
    
    obSurface = RootObData->surface;
    filterStack = RootObData->filterHead;
    while(filterStack != NULL) {
      filterStack->classData->renderFunc(obSurface, 
					 filterStack->surface, 0, 0,
					 filterStack->filterData);
      obSurface = filterStack->surface;
      filterStack = filterStack->next;
    } 

    SDL_BlitSurface(obSurface, NULL, FBSurface, NULL);
  } else {
    /* Just fill with colour. */

    SDL_FillRect(FBSurface, NULL, ClearColour);

    if(RootObData->filterHead) {
      /* Handle filters. */
      filterStack = RootObData->filterHead;
      obSurface = FBSurface;
      while(filterStack != NULL) {
	filterStack->classData->renderFunc(obSurface, 
					   filterStack->surface, 0, 0,
					   filterStack->filterData);
	obSurface = filterStack->surface;
	filterStack = filterStack->next;
      } 
      SDL_BlitSurface(obSurface, NULL, FBSurface, NULL);
    }
  }

  /* Draw each object in turn. */
  obData=RootObData->child;
  while(obData != NULL) {      
    if(! obData->activated) {
      /* Object not active. Skip. */
      obData = obData->child;
      continue;
    }
    obSurface=obData->surface;            

    if((obData->animLength != 0) && (obData->animSpeed != 0)) {
      animateOb(obData);
    }
    
    filterStack = obData->filterHead;
    if (filterStack != NULL) {
      /* Process filters*/
      
      /* No RLE acceleration as filters will be changing the surface
	 continuously anyway. */
      flags = SDL_SRCALPHA;
      while(filterStack != NULL) {
	filterStack->classData->renderFunc(obSurface, 
					   filterStack->surface, 0, 0,
					   filterStack->filterData);
	obSurface = filterStack->surface;
	filterStack = filterStack->next;
      } 
    } else {
      flags = SDL_SRCALPHA | SDL_RLEACCEL;
    }

    // :NOTE: Not optimal. Should set alpha once for last object.
    // and then copy if more filters are added or object changed.
    SDL_SetAlpha(obSurface, flags, obData->alpha);

    if ((obData->scaleWidth != -1) || (obData->scaleHeight != -1)) {
      drawScaledOb(FBSurface, obData);
    } else {
      destRect.x = obData->x;
      destRect.y = obData->y;
      SDL_BlitSurface(obSurface, NULL, FBSurface, &destRect);
    }

    if(obData->speed != 0) {
      if(moveOb(obData) == -1) {
	return -1;
      }
    }

    obData = obData->child;
  }
    
  //  SDL_UpdateRect(FBSurface, 0, 0, 0, 0);
  SDL_Flip(FBSurface);

#ifdef DEBUG
  if (frameCount == 100) {
    D3PRINTF(("draw: %ld milliseconds per frame (%ld fps)", 
	      (SDL_GetTicks() - LastCheckTime) / frameCount, 
	      1000/((SDL_GetTicks() - LastCheckTime) / frameCount)));
    LastCheckTime = SDL_GetTicks();
    frameCount = 0;
  } else {
    frameCount++;
  }
#endif

  return 1;
}  


/* Gets the current surface to be rendered to screen for object, f.ex. if there
   are filters, this could be the last filter surface. */

static SDL_Surface *getRenderingSurface(OilOb *ob) {
  if(ob->filterTail != NULL) {
    return ob->filterTail->surface;
  } else {
    return ob->surface;
  }
}


/* Draw a scaled object to the screen. */

static void drawScaledOb(SDL_Surface *renderTo, OilOb *ob) {
  int y, ySource, ySourceDec;
  int ySourceInc, ySourceDecInc;
  SDL_Surface *obSurface = getRenderingSurface(ob);
  SDL_Rect destRect, srcRect;

  if(ob->scaleWidth == -1) {
    /* Only do height stretch. */

    ySource = 0;
    ySourceDec = 0;
    ySourceInc = obSurface->h / ob->scaleHeight;
    ySourceDecInc = 1000 / ((float) ob->scaleHeight / 
			    (obSurface->h % ob->scaleHeight));
    srcRect.x = 0;
    srcRect.w = obSurface->w;
    srcRect.h = 1;
    destRect.x = ob->x;
    for(y=ob->y; y < (ob->y + ob->scaleHeight); y++) {
      srcRect.y = ySource;
      destRect.y = y;
      SDL_BlitSurface(obSurface, &srcRect, renderTo, &destRect);
      ySource += ySourceInc;
      ySourceDec += ySourceDecInc;
      if(ySourceDec >= 1000) {
	ySource++;
	ySourceDec -= 1000;
      }
    }
  }

  /* Don't care about the rest of the case for now. */

  return;
}


/**
 * Animate object according to its animation settings.
 */

static void animateOb(OilOb *oilData) {
  Uint32 diff = SDL_GetTicks() - oilData->lastFrameTime;
  int frame, frameAdd;

  frameAdd = diff / oilData->clicksPerFrame;
  if (! oilData->loop) {
    /* Not a loop. */
    frame = oilData->currentFrame + frameAdd;
    if(frame > oilData->animLength) {
      /* Animation finished. */      
      oilData->animSpeed = 0;
      oilData->surface = oilData->frames[oilData->animLength];
      oilData->currentFrame = oilData->animLength;
      return;
    }
  } else {
    frame = (oilData->currentFrame + frameAdd) 
      % oilData->animLength;
  }

  oilData->surface = oilData->frames[frame];
  oilData->currentFrame = frame;
  if (frameAdd > 0) {
    oilData->lastFrameTime = SDL_GetTicks();
  }

  return;
}


/**
 * Move object according to its movement settings.
 */

static int moveOb(OilOb *oilData) {
  unsigned long timeDiff;

  timeDiff = SDL_GetTicks() - oilData->moveStartTime;

  if(timeDiff >= oilData->moveTime) {
    /* Required time has elapsed. Move the object to the end
       and we're done. */
    oilData->x = oilData->toX;
    oilData->y = oilData->toY;
    oilData->speed = 0;
    if(oilData->moveCallback != NULL) {
      if(Tcl_EvalObjEx(oilData->interp, oilData->moveCallback, 
		       TCL_EVAL_GLOBAL) == TCL_ERROR) {	
	return -1;
      }
      
    }

    return 0;
  } else {
    oilData->x = oilData->fromX + (float) oilData->distanceX * 
      (((float) timeDiff) / oilData->moveTime);

    oilData->y = oilData->fromY + oilData->distanceY *
      (((float) timeDiff) / oilData->moveTime);    
  }

  return 0;
}


/** 
 * Deactivate object (do not render or handle any movements etc.).
 *
 * Usage: deactivate <object>
 */

static int deactivateCmd(ClientData clientData, Tcl_Interp *interp, int objc,
			 Tcl_Obj *const objv[]) {
  OilOb *oilData;

  if(objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "object");
    return TCL_ERROR;
  }

  if(Tcl_ConvertToType(interp, objv[1], &OilObType) == TCL_ERROR) {
    return TCL_ERROR;
  }
  oilData = objv[1]->internalRep.otherValuePtr;

  oilData->activated = 0;

  return TCL_OK;
}


/** 
 * Activate object.
 *
 * Usage: activate <object>
 */

static int activateCmd(ClientData clientData, Tcl_Interp *interp, int objc,
		       Tcl_Obj *const objv[]) {
  OilOb *oilData;

  if(objc < 2) {
    Tcl_WrongNumArgs(interp, 1, objv, "object");
    return TCL_ERROR;
  }

  if(Tcl_ConvertToType(interp, objv[1], &OilObType) == TCL_ERROR) {
    return TCL_ERROR;
  }
  oilData = objv[1]->internalRep.otherValuePtr;

  oilData->activated = 1;

  return TCL_OK;
}


/** 
 * Set speed of movement in pixels / minute.
 *
 * Usage: setMoveSpeed <object> <speed>
 */

static int setMoveSpeedCmd(ClientData clientData, Tcl_Interp *interp, int objc,
			   Tcl_Obj *const objv[]) {
  OilOb *oilData;
  int speed;

  if(objc < 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "object width");
    return TCL_ERROR;
  }

  if(Tcl_ConvertToType(interp, objv[1], &OilObType) == TCL_ERROR) {
    return TCL_ERROR;
  }
  oilData = objv[1]->internalRep.otherValuePtr;
  
  if(Tcl_GetIntFromObj(interp, objv[2], &speed) == TCL_ERROR) {
    return TCL_ERROR;
  }

  oilData->speed = speed;

  return TCL_OK;
}


/** 
 * Set scaling of object.
 *
 * Usage: scale <object> <width> <height>
 *
 * If either 'width' or 'height' is "original", that dimension of the 
 * object is returned to its original value.
 */

static int scaleCmd(ClientData clientData, Tcl_Interp *interp, int objc,
			   Tcl_Obj *const objv[]) {
  OilOb *oilData;
  int width, height;

  if(objc < 4) {
    Tcl_WrongNumArgs(interp, 1, objv, "object width height");
    return TCL_ERROR;
  }

  if(Tcl_ConvertToType(interp, objv[1], &OilObType) == TCL_ERROR) {
    return TCL_ERROR;
  }
  oilData = objv[1]->internalRep.otherValuePtr;

  /* Get width argument. */
  if(Tcl_GetIntFromObj(interp, objv[2], &width) == TCL_ERROR) {
    if (strcmp(Tcl_GetStringFromObj(objv[2], NULL), "original") == 0) {
      Tcl_ResetResult(interp);
      oilData->scaleWidth = -1;
    } else {
      return TCL_ERROR;
    }
  } else {
    if(width < 1) {
      Tcl_SetObjResult(interp, Tcl_NewStringObj("Width out of range.",
						-1));
      return TCL_ERROR;
    }

    oilData->scaleWidth = width;
  }

  /* Get height. */
  if(Tcl_GetIntFromObj(interp, objv[3], &height) == TCL_ERROR) {
    if (strcmp(Tcl_GetStringFromObj(objv[3], NULL), "original") == 0) {
      Tcl_ResetResult(interp);
      oilData->scaleHeight = -1;
    } else {
      return TCL_ERROR;
    }
  } else {
    if(height < 1) {
      Tcl_SetObjResult(interp, Tcl_NewStringObj("Height out of range.",
						-1));
      return TCL_ERROR;
    }

    oilData->scaleHeight = height;
  }

  return TCL_OK;
}


/** 
 * Set object's colour key.
 *
 * Usage: setColourKey <object> <r> <g> <b>
 * Usage: setColourKey <object> none
 *
 * If the latter form is used, the colour key is reset. 
 */

static int setColourKeyCmd(ClientData clientData, Tcl_Interp *interp, int objc,
			   Tcl_Obj *const objv[]) {
  OilOb *oilData;
  int r, g, b, i;
  Uint32 colourKey, flag;
  FilterInstance *filterStack;

  if (objc < 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "object [r g b|none]");
    return TCL_ERROR;
  }
  
  if(Tcl_ConvertToType(interp, objv[1], &OilObType) == TCL_ERROR) {
    return TCL_ERROR;
  }
  oilData = objv[1]->internalRep.otherValuePtr;

  if (objc == 3) {
    /* Expecting 'none' as argument. */
    
    if (strcmp(Tcl_GetStringFromObj(objv[2], NULL), "none") == 0) {
      /* Reset colour key. */
      flag = 0;
      r = 0;
      g = 0;
      b = 0;
    } else {
      Tcl_SetObjResult(interp, Tcl_NewStringObj("Not RGB value or 'none'.",
						-1));
      return TCL_ERROR;
    }
  } else {
    /* R G B values given. */

    if(objc < 5) {
      Tcl_WrongNumArgs(interp, 1, objv, "object [r g b|none]");
      return TCL_ERROR;
    }
    
    D3PRINTF(("setColourKeyCmd: oilData->name: %s", oilData->name));
    if(Tcl_GetIntFromObj(interp, objv[2], &r) == TCL_ERROR) {
      return TCL_ERROR;
    }
    if(Tcl_GetIntFromObj(interp, objv[3], &g) == TCL_ERROR) {
      return TCL_ERROR;
    }
    if(Tcl_GetIntFromObj(interp, objv[4], &b) == TCL_ERROR) {
      return TCL_ERROR;
    }
    
    if(oilData->filterHead == NULL) {
      flag = SDL_SRCCOLORKEY|SDL_RLEACCEL;
    } else {
      /* For some reason RLE acceleration doesn't work with the filter 
	 surfaces.
	 Perhaps blitting from one surface to another (instead of framebuffer)
	 is problematic with RLE? Anyhow, RLE is probably quite useless with
	 filters on, as they keep changing the result anyway. */

      flag = SDL_SRCCOLORKEY;
    }
  }

  filterStack = oilData->filterHead;
  if(! oilData->filterTail) {
    /* We only want to set the colour key for the non-filter surfaces
       if filters have not been set (if filters are set, the surface
       data will be used in them and we do not want colour keys to 
       affect that). */
    colourKey = SDL_MapRGB(oilData->surface->format, r, g, b);
    SDL_SetColorKey(oilData->surface, flag, colourKey);
    
    /* Set colour key for animation frames. */
    if(oilData->frames != NULL) {
      for (i=0; i < oilData->animLength; i++) {
	SDL_SetColorKey(oilData->frames[i], flag, colourKey);
      }
    }
  } else {
    /* Filters set. */

    /*
    while(filterStack != NULL) {
    colourKey = SDL_MapRGB(filterStack->surface->format, r, g, b);
      SDL_SetColorKey(filterStack->surface, SDL_SRCCOLORKEY, colourKey);
      filterStack = filterStack->next;
    } 
    */
    colourKey = SDL_MapRGB(oilData->filterTail->surface->format, r, g, b);
    SDL_SetColorKey(oilData->filterTail->surface, flag, colourKey);
  }

  return TCL_OK;
}


/** 
 * Set object position.
 *
 * Usage: setPos <object> <x> <y>
 */

static int setPosCmd(ClientData clientData, Tcl_Interp *interp, int objc,
		     Tcl_Obj *const objv[]) {
  OilOb *oilData;
  int x, y, distanceX, distanceY, distance;

  if(objc < 4) {
    Tcl_WrongNumArgs(interp, 1, objv, "object x y");
    return TCL_ERROR;
  }

  if(Tcl_ConvertToType(interp, objv[1], &OilObType) == TCL_ERROR) {
    return TCL_ERROR;
  }
  oilData = objv[1]->internalRep.otherValuePtr;

  if(Tcl_GetIntFromObj(interp, objv[2], &x) == TCL_ERROR) {
    return TCL_ERROR;
  }
  if(Tcl_GetIntFromObj(interp, objv[3], &y) == TCL_ERROR) {
    return TCL_ERROR;
  }

  oilData->x = x;
  oilData->y = y;

  if (oilData->speed) {
    /* Movement was set. Update its information to reflect new position. */
    oilData->fromX = x;
    oilData->fromY = y;
    distanceX = abs(oilData->toX - oilData->x);
    distanceY = abs(oilData->toY - oilData->y);
    if (distanceX > distanceY) {
      distance = distanceX;
    } else {
      distance = distanceY;
    }
    oilData->distanceX = oilData->toX - oilData->x;
    oilData->distanceY = oilData->toY - oilData->y;
    oilData->moveTime = (((float) distance / oilData->speed) * 1000);
  }


  return TCL_OK;
}


/** 
 * Command used to specify object animation based on a list of bitmap files.
 * Usage: setAnimSpeed <object> <speed>
 *
 * 'object'  Oil object.
 * 'speed'  Speed of animation in frames per second.
 */

static int setAnimSpeedCmd(ClientData clientData, Tcl_Interp *interp, 
			   int objc, Tcl_Obj *const objv[]) {
  OilOb *oilData;
  int speed;  

  if(objc<3) {
    Tcl_WrongNumArgs(interp, 1, objv, "object speed");
    return TCL_ERROR;
  }

  if(Tcl_ConvertToType(interp, objv[1], &OilObType) == TCL_ERROR) {
    return TCL_ERROR;
  }
  oilData = objv[1]->internalRep.otherValuePtr;

  if(Tcl_GetIntFromObj(interp, objv[2], &speed) == TCL_ERROR) {
    return TCL_ERROR;
  }

  oilData->clicksPerFrame = 1000 / (((float) speed) / 60);
  oilData->lastFrameTime = SDL_GetTicks();
  oilData->animSpeed = speed;

  return TCL_OK;
}


/**
 * Command: Copies source object's animation frames to another object.
 *
 * Usage: copyFrames <srcObject> <destObject>
 */

static int copyFramesCmd(ClientData clientData, Tcl_Interp *interp, 
			 int objc, Tcl_Obj *const objv[]) {
  OilOb *obDataSrc, *obDataDest;  
  int i;
  SDL_Surface *newSurface;
  Uint32 colorKey, flags;
  
  if(objc<3) {
    Tcl_WrongNumArgs(interp, 1, objv, "srcObject destObject");
    return TCL_ERROR;
  }

  if(Tcl_ConvertToType(interp, objv[1], &OilObType) == TCL_ERROR) {
    return TCL_ERROR;
  }
  obDataSrc = objv[1]->internalRep.otherValuePtr;

  if(Tcl_ConvertToType(interp, objv[2], &OilObType) == TCL_ERROR) {
    return TCL_ERROR;
  }
  obDataDest = objv[2]->internalRep.otherValuePtr;

  /* First free old animation frames. */
  if (obDataDest->animLength > 0) {
    for (i=0; i < obDataDest->animLength; i++) {
      SDL_FreeSurface(obDataDest->frames[i]);
    }

    obDataDest->frames = realloc(obDataDest->frames, 
			      sizeof(SDL_Surface *) * obDataSrc->animLength);
  } else {
    obDataDest->frames = malloc(sizeof(SDL_Surface *) * obDataSrc->animLength);
  }

  obDataDest->animLength = obDataSrc->animLength;

  /* Copy source frames over. */
  for(i=0; i < obDataSrc->animLength; i++) {
    // :NOTE: Bug: we do not check if surface creation was successful.
    newSurface = Oil_CreateDisplaySurface16(obDataSrc->surface->w, 
					    obDataSrc->surface->h);    

    /* Temporarily unset colour key for blitting to new surface. */
    flags = obDataSrc->frames[i]->flags;
    colorKey = obDataSrc->frames[i]->format->colorkey;
    SDL_SetColorKey(obDataSrc->frames[i], 0, 0);

    SDL_BlitSurface(obDataSrc->frames[i], NULL, newSurface, NULL);
    obDataDest->frames[i] = newSurface;

    SDL_SetColorKey(obDataSrc->frames[i], flags, colorKey);
  }

  obDataDest->loop = obDataSrc->loop;
  obDataDest->clicksPerFrame = obDataSrc->clicksPerFrame;
  obDataDest->currentFrame = 0;
  obDataDest->lastFrameTime = SDL_GetTicks();
  obDataDest->animSpeed = obDataSrc->animSpeed;
  obDataDest->surface = obDataDest->frames[0];

  return TCL_OK;
}


/** 
 * Command used to specify object animation based on a list of bitmap files.
 * Usage: loadAnimFrames <object> <speed> <loop> <imageFileList>
 *
 * 'object'  Oil object.
 * 'speed'  Speed of animation in frames per second.
 * 'loop'  True if we are to loop and false if not. 
 * 'imageFileList'  List of files to load in as animation.
 */

static int loadAnimFramesCmd(ClientData clientData, Tcl_Interp *interp, 
			     int objc,
			     Tcl_Obj *const objv[]) {
  OilOb *oilData;
  int newAnimLength, i, speed, loop;
  Tcl_Obj **fileList;
  char *fileName;
  SDL_Surface *bmpSurface, *tmpSurface;
  

  if(objc<5) {
    Tcl_WrongNumArgs(interp, 1, objv, "object speed loop imageFileList");
    return TCL_ERROR;
  }

  if(Tcl_ConvertToType(interp, objv[1], &OilObType) == TCL_ERROR) {
    return TCL_ERROR;
  }
  oilData = objv[1]->internalRep.otherValuePtr;

  if(Tcl_GetIntFromObj(interp, objv[2], &speed) == TCL_ERROR) {
    return TCL_ERROR;
  }
  if(Tcl_GetBooleanFromObj(interp, objv[3], &loop) == TCL_ERROR) {
    return TCL_ERROR;
  }

  if(Tcl_ListObjGetElements(interp, objv[4], &newAnimLength, &fileList) 
     == TCL_ERROR) {
    return TCL_ERROR;
  }

  /* First free old animation frames. */
  if (oilData->animLength > 0) {
    for (i=0; i < oilData->animLength; i++) {
      SDL_FreeSurface(oilData->frames[i]);
    }

    oilData->frames = realloc(oilData->frames, 
			      sizeof(SDL_Surface *) * newAnimLength);
  } else {
    oilData->frames = malloc(sizeof(SDL_Surface *) * newAnimLength);
  }

  oilData->animLength = newAnimLength;
  for(i=0; i < newAnimLength; i++) {
    if((fileName=Tcl_GetStringFromObj(fileList[i], NULL))
       == NULL) {
      Tcl_SetObjResult(interp, Tcl_NewStringObj("Invalid file path argument.",
						-1));
      return TCL_ERROR;
    }

    if((bmpSurface=SDL_LoadBMP(fileName)) == NULL) {
      Tcl_SetObjResult(interp, Tcl_NewStringObj(SDL_GetError(), -1));
      return TCL_ERROR;
    }

    tmpSurface=SDL_DisplayFormat(bmpSurface);
    SDL_FreeSurface(bmpSurface);
    if(tmpSurface == NULL) {
      Tcl_SetObjResult(interp, Tcl_NewStringObj(SDL_GetError(), -1));
      return TCL_ERROR;
    }

    oilData->frames[i] = tmpSurface;
  }

  oilData->lastFrameTime = SDL_GetTicks();
  SDL_FreeSurface(oilData->surface);
  oilData->surface = oilData->frames[0];
  oilData->currentFrame = 0;
  D5PRINTF(("loadAnimFramesCmd: speed: %d", speed));
  D4PRINTF(("loadAnimFramesCmd: speed/60: %f", ((float) speed)/60));
  oilData->clicksPerFrame = 1000 / (((float) speed) / 60);
  oilData->loop = loop;
  oilData->animSpeed = speed;

  D3PRINTF(("loadAnimFramesCmd: clicksPerFrame: %ld", 
	    oilData->clicksPerFrame));
  D3PRINTF(("loadAnimFramesCmd: loop: %d", oilData->loop));
  D3PRINTF(("loadAnimFramesCmd: oilData->animSpeed: %d", oilData->animSpeed));
  D3PRINTF(("loadAnimFramesCmd: animLength: %d", oilData->animLength));

  return TCL_OK;
}


/** 
 * Command used to specify a moving object. 
 * Usage: move <object> <toX> <toY> <speed> [callback]
 *
 * 'toX'  X co-ordinate to move towards.
 * 'toY'  Y co-ordinate to move towards.
 * 'speed'  Speed at which to move, in pixels per second.
 * 'callback'  Script to call when movement is finished.
 */

static int moveCmd(ClientData clientData, Tcl_Interp *interp, int objc,
		   Tcl_Obj *const objv[]) {
  OilOb *oilData;
  int toX, toY, speed, distanceX, distanceY, distance;

  if(objc<5) {
    Tcl_WrongNumArgs(interp, 1, objv, "object toX toY speed ?callback?");
    return TCL_ERROR;
  }

  if(Tcl_ConvertToType(interp, objv[1], &OilObType) == TCL_ERROR) {
    return TCL_ERROR;
  }
  oilData = objv[1]->internalRep.otherValuePtr;

  if(Tcl_GetIntFromObj(interp, objv[2], &toX) == TCL_ERROR) {
    return TCL_ERROR;
  }
  if(Tcl_GetIntFromObj(interp, objv[3], &toY) == TCL_ERROR) {
    return TCL_ERROR;
  }
  if(Tcl_GetIntFromObj(interp, objv[4], &speed) == TCL_ERROR) {
    return TCL_ERROR;
  }

  if(oilData->moveCallback != NULL) {
    /* Previous callback script exists. */
    Tcl_DecrRefCount(oilData->moveCallback);
  }
  if (objc == 6) {
    /* Callback script given. */

    oilData->moveCallback = objv[5];
    Tcl_IncrRefCount(objv[5]);
  } else {
    oilData->moveCallback = NULL;
  }

  oilData->moveStartTime = SDL_GetTicks();
  oilData->fromX = oilData->x;
  oilData->fromY = oilData->y;
  oilData->toX = toX;
  oilData->toY = toY;
  oilData->speed = speed;
  distanceX = abs(toX - oilData->x);
  distanceY = abs(toY - oilData->y);
  if (distanceX > distanceY) {
    distance = distanceX;
  } else {
    distance = distanceY;
  }
  oilData->distanceX = toX - oilData->x;
  oilData->distanceY = toY - oilData->y;
  D3PRINTF(("moveCmd: distance: %d, speed: %d", distance, speed));
  oilData->moveTime = (((float) distance / speed) * 1000);
  D3PRINTF(("moveCmd: moveTime: %ld", oilData->moveTime));

  return TCL_OK;
}


/** Forces a redraw of the screen.
 *  Usage: draw
 */

static int drawCmd(ClientData clientData, Tcl_Interp *interp, int objc,
		   Tcl_Obj *const objv[]) {
  draw();
  return TCL_OK;
}


/** 
 * Set frame of object to display.
 *
 *  Usage: setParent <object> <frame>
 */

static int setFrameCmd(ClientData clientData, Tcl_Interp *interp, int objc,
		       Tcl_Obj *const objv[]) {
  OilOb *oilData;
  int frame;

  if(objc<3) {
    Tcl_WrongNumArgs(interp, 1, objv, "object newParentObj");
    return TCL_ERROR;
  }

  if(Tcl_ConvertToType(interp, objv[1], &OilObType) == TCL_ERROR) {
    return TCL_ERROR;
  }
  oilData = objv[1]->internalRep.otherValuePtr;

  if(Tcl_GetIntFromObj(interp, objv[2], &frame) == TCL_ERROR) {
    return TCL_ERROR;
  }

  if(frame >= oilData->animLength) {
    Tcl_SetObjResult(interp, Tcl_NewStringObj("Frame out of range", -1));
    return TCL_ERROR;
  }

  oilData->surface = oilData->frames[frame];

  return TCL_OK;
}


/** Changes parent of object, and thus the display order.
 *  Usage: setParent <object> <newParentObj>
 *  'object'  Oil object.
 *  'newParentObj'  Handle to oil object which should be the new parent.
 */

static int setParentCmd(ClientData clientData, Tcl_Interp *interp, int objc,
			Tcl_Obj *const objv[]) {
  OilOb *oilData;
  OilOb *newParentData;

  if(objc<3) {
    Tcl_WrongNumArgs(interp, 1, objv, "object newParentObj");
    return TCL_ERROR;
  }

  if(Tcl_ConvertToType(interp, objv[1], &OilObType) == TCL_ERROR) {
    return TCL_ERROR;
  }
  oilData = objv[1]->internalRep.otherValuePtr;

  if(Tcl_ConvertToType(interp, objv[2], &OilObType) == TCL_ERROR) {
    return TCL_ERROR;
  }
  newParentData = objv[2]->internalRep.otherValuePtr;

  if(oilData->parent != NULL) {
    oilData->parent->child = oilData->child;
  }
  if(oilData->child) {
    oilData->child->parent = oilData->parent;
  }
  oilData->parent = newParentData;
  oilData->child = newParentData->child;
  if(newParentData->child != NULL) {
    newParentData->child->parent = oilData;
  }
  newParentData->child = oilData;

  return TCL_OK;
}


/** Adds filter to object. 
 *  Usage: filter <object> <filterName>
 */

static int filterCmd(ClientData clientData, Tcl_Interp *interp, int objc,
		      Tcl_Obj *const objv[]) {
  OilOb *oilData;
  Oil_Filter *filter;
  char *filterName;
  SDL_Surface *filterSurface, *srcSurface;
  void *filterData;
  FilterInstance *newInstance;

  if(objc<3) {
    Tcl_WrongNumArgs(interp, 1, objv, "object filter ?options?");
    return TCL_ERROR;
  }

  if(Tcl_ConvertToType(interp, objv[1], &OilObType) == TCL_ERROR) {
    return TCL_ERROR;
  }
  oilData = objv[1]->internalRep.otherValuePtr;

  filterName = Tcl_GetString(objv[2]);
  if((filter = Oil_FindFilter(filterName)) == NULL) {
    Tcl_SetObjResult(interp, Tcl_NewStringObj("Could not find filter.",
					      -1));
    return TCL_ERROR;
  }

  D3PRINTF(("filterCmd: oilData->name: %s, filterName: %s", oilData->name, 
	    filterName));

  filterData = NULL;

  if(oilData->filterTail != NULL) {
    /* Previous filter exists. */
    srcSurface = oilData->filterTail->surface;
    if((filterSurface = filter->createFunc(srcSurface, NULL, &filterData)) 
       == NULL) {
      Tcl_SetObjResult(interp, Tcl_NewStringObj("Unable to create filter.",
						-1));
      return TCL_ERROR;
    }

    newInstance = filterInstanceNew(filter, filterSurface, filterData);
    oilData->filterTail->next = newInstance;
    oilData->filterTail = newInstance;
  } else {
    /* No previous filter. Use object's surface as source. */
    srcSurface = oilData->surface;
    if((filterSurface = filter->createFunc(srcSurface, NULL, &filterData)) 
       == NULL) {
      Tcl_SetObjResult(interp, Tcl_NewStringObj("Unable to create filter.",
						-1));
      return TCL_ERROR;
    }

    oilData->filterHead = oilData->filterTail =
      filterInstanceNew(filter, filterSurface, filterData);
  }

  if (srcSurface->flags & SDL_SRCCOLORKEY) {
    /* Move colour key to new surface. */
    SDL_SetColorKey(filterSurface, SDL_SRCCOLORKEY,
		    srcSurface->format->colorkey);
    SDL_SetColorKey(srcSurface, 0, 0);
  } 

  return TCL_OK;
}


/** Configure object's alpha value.
 *
 *  Usage: setAlpha <object> <alpha>
 *  'object'   Oil object to load bitmap into.
 *  'alpha'  Alpha value in range 0-255
*/

static int setAlphaCmd(ClientData clientData, Tcl_Interp *interp, int objc,
		       Tcl_Obj *const objv[]) {
  OilOb *oilData;
  //  Uint8 alpha;
  int alphaParam;

  if(objc < 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "object alpha");
    return TCL_ERROR;
  }
   
  if(Tcl_ConvertToType(interp, objv[1], &OilObType) == TCL_ERROR) {
    return TCL_ERROR;
  }
  oilData = objv[1]->internalRep.otherValuePtr;

  if(Tcl_GetIntFromObj(interp, objv[2], &alphaParam) == TCL_ERROR) {
    return TCL_ERROR;
  }
  if(alphaParam < 0 || alphaParam > 255) {
    Tcl_SetObjResult(interp, Tcl_NewStringObj("Alpha out of range.",
					      -1));
    return TCL_ERROR;
  }

  oilData->alpha = alphaParam;
  //  SDL_SetAlpha(oilData->surface, SDL_SRCALPHA, alphaParam);

  return TCL_OK;
}


/** Loads and draws a bitmap to the object. Supported file format: BMP.
 *
 *  Usage: loadBitmap <object> <file>
 *  'object'   Oil object to load bitmap into.
 *  'file'  Path to BMP file.
*/

static int loadBitmap(ClientData clientData, Tcl_Interp *interp, int objc,
		      Tcl_Obj *const objv[]) {
  SDL_Surface *tmpSurface, *bmpSurface;
  char *fileName;
  OilOb *oilData;
  int i;

  D1PRINTF(("loadBitmap called"));
  D3PRINTF(("loadBitmap: objc: %d", objc));
  if(objc < 3) {
    Tcl_WrongNumArgs(interp, 1, objv, "object filename");
    return TCL_ERROR;
  }
   
  D2PRINTF(("type conversion"));
  if(Tcl_ConvertToType(interp, objv[1], &OilObType) == TCL_ERROR) {
    return TCL_ERROR;
  }
  oilData = objv[1]->internalRep.otherValuePtr;
    
  if((fileName=Tcl_GetStringFromObj(objv[2], NULL))
     == NULL) {
    Tcl_SetObjResult(interp, Tcl_NewStringObj("Invalid file path argument.",
					      -1));
    return TCL_ERROR;
  }

  if((bmpSurface=SDL_LoadBMP(fileName)) == NULL) {
    Tcl_SetObjResult(interp, Tcl_NewStringObj(SDL_GetError(), -1));
    return TCL_ERROR;
  }

  tmpSurface=SDL_DisplayFormat(bmpSurface);
  SDL_FreeSurface(bmpSurface);
  if(tmpSurface == NULL) {
    Tcl_SetObjResult(interp, Tcl_NewStringObj(SDL_GetError(), -1));
    return TCL_ERROR;
  }

  oilData = getOilData(objv[1]);
  
  /* Remove old data, if it exists. */
  if(oilData->animLength != 0) {
    /* Free animation frames. */
    for (i=0; i < oilData->animLength; i++) {
      SDL_FreeSurface(oilData->frames[i]);
    }
    free(oilData->frames);
    oilData->animLength = 0;
    oilData->animSpeed = 0;
    oilData->loop = 0;
  } else {
    if(oilData->surface != FBSurface) {
      SDL_FreeSurface(oilData->surface);
    } 
  }

  if(oilData->root) {
    /* Root object. We need to create a surface equal to the size of the
       root, and blit the bitmap there. */
    if((bmpSurface = Oil_CreateDisplaySurface16(FBSurface->w, FBSurface->h))
       == NULL) {
      SDL_FreeSurface(tmpSurface);
      Tcl_SetObjResult(interp, Tcl_NewStringObj(SDL_GetError(), -1));
      return TCL_ERROR;
    }
    
    // SDL_SetAlpha(tmpSurface, 0, 255);
    SDL_BlitSurface(tmpSurface, NULL, bmpSurface, NULL);
    //    SDL_SetAlpha(bmpSurface, 0, 255);
    SDL_FreeSurface(tmpSurface);
    tmpSurface=bmpSurface;
  }
    

  oilData->surface = tmpSurface;
  D5PRINTF(("loadBitmap: oilData->surface->flags: %x", 
	    oilData->surface->flags));

  return TCL_OK;
}
  
    
/** Implements command to set up a frame buffer. 
 *    
 * Usage: setVideo <width> <height>
 *
 * Return: identifier of root object. Recommended practise is to
 * set to a variable to avoid unnecessary Tcl_Obj destruction and 
 * re-creation.
 */

int setVideo(ClientData clientData, Tcl_Interp *interp, int objc, 
	     Tcl_Obj *const objv[]) {
  int width, height;
  Tcl_Obj *rootOb;
  char *root;

  if(objc < 3) {
    Tcl_SetObjResult(interp, Tcl_NewStringObj("Too few arguments.", 
					      -1));
    return TCL_ERROR;
  }
  if(Tcl_GetIntFromObj(interp, objv[1], &width) == TCL_ERROR) {
    return TCL_ERROR;
  }
  if(Tcl_GetIntFromObj(interp, objv[2], &height) == TCL_ERROR) {
    return TCL_ERROR;
  }
  // FBSurface=SDL_SetVideoMode(width, height, 16, SDL_SWSURFACE|SDL_DOUBLEBUF|SDL_FULLSCREEN);
  FBSurface=SDL_SetVideoMode(width, height, 16, SDL_SWSURFACE|SDL_DOUBLEBUF);

  D5PRINTF(("setVideo: sizeof(ROOT_NAME): %d", sizeof(ROOT_NAME)));
  root=malloc(sizeof(ROOT_NAME));
  strcpy(root, ROOT_NAME);
  if((rootOb=oilOb_New(NULL, root, interp, FBSurface, 0, 0)) == NULL) {
    Tcl_SetObjResult(interp, Tcl_NewStringObj(SDL_GetError(), -1));
    return TCL_ERROR;
  }

  ClearColour=SDL_MapRGB(FBSurface->format , 0, 0, 0);
  RootObData = rootOb->internalRep.otherValuePtr;
  RootObData->root = 1;

  Tcl_SetObjResult(interp, rootOb);
  return TCL_OK;
} 


/** Implements Tcl command to list all Oil objects. */

int listObs(ClientData clientData, Tcl_Interp *interp, int objc,
	    Tcl_Obj *const objv[]) {
  Tcl_Obj *obList;
  Tcl_HashSearch searchTracker;
  Tcl_HashEntry *hashEntry;  

  hashEntry=Tcl_FirstHashEntry(&Str2OilMap, &searchTracker);
  if(hashEntry != NULL) {
    obList = Tcl_NewListObj(0, NULL);
    Tcl_ListObjAppendElement(interp, obList, 
			     Tcl_NewStringObj(Tcl_GetHashKey(&Str2OilMap, 
							     hashEntry), -1));

    while((hashEntry=Tcl_NextHashEntry(&searchTracker)) 
	    != NULL) {
      Tcl_ListObjAppendElement(interp, obList, 
			       Tcl_NewStringObj(Tcl_GetHashKey(&Str2OilMap, 
							       hashEntry), 
						-1));

    }
    Tcl_SetObjResult(interp, obList);
  } 

  return TCL_OK;
}


void mainLoop(void) {

  while(1) {
    //    Tcl_ServiceAll();
    Tcl_DoOneEvent(TCL_DONT_WAIT);
    draw();
  }
}
